import re

import numpy as np
import pandas as pd
import logging

logger = logging.getLogger(__name__)

header = [
    "qseqid", "sseqid", "pident", "length", "evalue", "bitscore", "slen", "stitle", "qlen"
]


class Annotate(object):
    def __init__(self, tmp_blast_file):
        df = pd.read_csv(tmp_blast_file, delimiter="\t", header=None)
        df.columns = header

        # df = df.head(10)

        self.genes_list = list(df["qseqid"].unique())
        self.df = df

    def get_df_for_spcific_gene(self, gene):
        return self.df[self.df["qseqid"] == gene]

    @staticmethod
    def get_df_for_genes_same_length_range(df):
        """
        qlen is the size of the protein we want to annotate.
        slen is the size of the protein in the database.
        """
        offset = int(df["qlen"].unique()[0] * 0.10)

        _df = df[df["slen"] > df["qlen"] - offset]
        _df = _df[_df["slen"] < _df["qlen"] + offset]
        return _df

    @staticmethod
    def remove_species(stitle):
        pattern = r"\[.*?\]"
        return re.sub(pattern, "", stitle).rstrip(" ")

    def default_pass(self, _df, step):
        for index, row in _df.iterrows():
            if row["stitle"] is np.NAN:
                # in case some database entry don't contain a description/product, we skip it.
                logger.warning(f"Missing subject title for blast database entry: {row['sseqid']}")
                continue

            blast_result = row.to_dict()
            blast_result["stitle"] = self.remove_species(blast_result["stitle"])
            logging.debug(f"{step}: {blast_result}")
            return blast_result
        return None

    def run(self):
        qualifiers = []
        for gene in self.genes_list:
            _df_all = self.get_df_for_spcific_gene(gene)
            _df_same_length = self.get_df_for_genes_same_length_range(_df_all)

            blast_result = self.default_pass(_df_same_length, "1st")
            if not blast_result:
                blast_result = self.default_pass(_df_all, "2nd")

            if blast_result:
                qualifiers.append(blast_result)
            else:
                logger.info(f"No good candidate found for CDS #{gene}")

        return qualifiers



import subprocess
from collections import defaultdict
import logging
import os
import sys

BLAST_FMT = "qseqid sseqid pident length evalue bitscore slen stitle qlen"
logger = logging.getLogger(__name__)


def run_blast(threads, query_file, blast_file, database):
    if not os.path.exists(database):
        logging.error("Default database not found.")
        sys.exit(1)

    blast_cmd = [
        "blastp",
        "-db", database,
        "-query", query_file,
        "-evalue", "1e-5",
        "-outfmt", f"'6 {BLAST_FMT}'",
        "-num_threads", f"{threads}",
        "-task", "blastp-fast",
        "-out", blast_file
    ]

    try:
        rt = subprocess.Popen(" ".join(blast_cmd), shell=True)
        rt.communicate()

    except FileNotFoundError:
        logging.error("blast not found!")
        sys.exit(1)


def run_prodigal(path):
    """
    Expects as input a FASTA file and return the prodigal default output as a list:
    ['>1_70_930_+', '>2_1282_2259_+', '>3_2276_2689_+', '>4_3116_3667_-', ... ]
    """
    prodigal_cmd = f"prodigal -i {path} -f sco -p meta"
    try:
        rt = subprocess.Popen(prodigal_cmd.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except FileNotFoundError:
        logging.error("Prodigal not found!")
        sys.exit(1)

    output, error = rt.communicate()
    output = output.decode("utf-8")
    features = defaultdict(list)

    idx = 1
    for line in output.split("\n"):
        if "seqhdr" in line:
            contig = line.split('"')[1].split(" ")[0]
        if line.startswith(">"):
            orf = line.split("_")
            orf[0] = f">{idx}"
            idx += 1
            features[contig].append("_".join(orf))

    return features


def run_phanotate(path):
    phanotate_cmd = f"phanotate.py {path}"
    rt = os.popen(phanotate_cmd).read()
    rt = rt.split("\n")

    features = defaultdict(list)
    idx = 1
    for line in rt:
        if line.split("\t") == [""]:
            break

        if not line.startswith("#"):
            orf = line.split("\t")[0:4]
            strand = orf[2]
            contig = orf[3]
            start, end = sorted([int(x) for x in orf[0:2]])
            orf = "_".join([f">{idx}", str(start), str(end), strand])
            idx += 1
            features[contig].append(orf)

    return features


def run_trnascan(path):
    """
    """

    trnascan_cmd = f"tRNAscan-SE -B -q --brief {path}"
    rt = os.popen(trnascan_cmd).read()
    rt = rt.split("\n")

    features = defaultdict(list)

    for line in rt[:-1]:
        trna = line.split("\t")
        orf = "_".join([f">{trna[1]}", str(trna[2].strip()), str(trna[3].strip()), trna[4].strip(), trna[5].strip()])
        features[trna[0].strip()].append(orf)

    return features



import datetime
import logging
import os
import re
import shutil
import sys
import tempfile
from Bio import SeqIO
from Bio.SeqFeature import SeqFeature, FeatureLocation
from src import tools
from src.annotate import Annotate
from src.core import cds_calling_from_genbank, get_database_default_path, validate_input
from src.output import write_gbk

logger = logging.getLogger(__name__)
total_steps = 9


class Pipeline(object):
    def __init__(self, database: str, input_file: str, prodigal: bool, threads: int, output_file: str, tmp_dir: str,
                 merge_gbk: bool, locus_tag: str):

        self.database = database if database else str(get_database_default_path())
        self.contig_file = input_file
        self.locus_tag = locus_tag

        # Operational options
        self.use_prodigal = prodigal
        self.blast_threads = threads
        self.output_file = output_file
        self.merge_gbk = merge_gbk

        tmp_dir_handler = tempfile.TemporaryDirectory()
        self.tmp_dir = tmp_dir if tmp_dir else tmp_dir_handler.name
        if tmp_dir:
            os.mkdir(tmp_dir)

        self.tmp_input_gbk_file = os.path.join(self.tmp_dir, 'input.gbk')  # genbank
        self.tmp_input_fna_file = os.path.join(self.tmp_dir, 'input.fna')  # contigs
        self.tmp_query_faa_file = os.path.join(self.tmp_dir, 'query.faa')  # ORFs
        self.tmp_blast_tsv_file = os.path.join(self.tmp_dir, 'blast.tsv')  # Blast result

        self.genome = None
        self.orf_map = None
        self.trna_map = None
        self.run()

    def run(self):
        self.load_genome_from_file()
        self.add_contig_metadata()

        logger.info(f"[2/{total_steps}] Starting ORF calling")
        self.orf_map = self.orf_calling()

        logger.info(f"[3/{total_steps}] Starting tRNA calling")
        self.trna_map = self.rna_calling()

        logger.info(f"[4/{total_steps}] Loading features into biopython SeqFeature")
        self.load_features()

        logger.info(f"[5/{total_steps}] Preparing blast query file")
        self.prepare_query_file()

        logger.info(f"[6/{total_steps}] Running blast with {self.database}")
        tools.run_blast(self.blast_threads, self.tmp_query_faa_file, self.tmp_blast_tsv_file, self.database)

        logger.info(f"[7/{total_steps}] Selecting best hits")
        qualifiers = {record["qseqid"]: record for record in Annotate(self.tmp_blast_tsv_file).run()}

        logger.info(f"[8/{total_steps}] Enriching features")
        self.enrich_features(qualifiers)

        logger.info(f"[9/{total_steps}] Writing output.")
        write_gbk(self.genome.values(), self.output_file)

    def load_genome_from_file(self):
        input_type = validate_input(self.contig_file)
        if not input_type:
            logger.error("Failed to detect input type")
            sys.exit(1)

        logger.info(f"[1/{total_steps}] Input type inferred as {input_type}")

        if input_type == 'FASTA' and self.merge_gbk:
            logger.error("Can't disable ORF calling if input is a FASTA file.")
            sys.exit(1)

        if input_type == 'FASTA':
            shutil.copyfile(self.contig_file, self.tmp_input_fna_file)
        elif input_type == 'GENBANK':
            with open(self.contig_file, "r") as input_handle:
                sequences = SeqIO.parse(input_handle, "genbank")
                with open(self.tmp_input_fna_file, "w") as output_handle:
                    SeqIO.write(sequences, output_handle, "fasta")
            shutil.copyfile(self.contig_file, self.tmp_input_gbk_file)
        else:
            logger.error("The code should never have hit this else!")
            sys.exit(1)

        self.contig_file = self.tmp_input_fna_file
        try:
            self.genome = SeqIO.to_dict(SeqIO.parse(self.contig_file, "fasta"))
        except ValueError as e:
            duplicated_contig = re.findall(r"'(.*?)'", str(e), re.DOTALL)[0]
            logger.error(f"A duplicated record was found: {duplicated_contig}")
            sys.exit(1)

    def add_contig_metadata(self):
        today_date = str(datetime.date.today().strftime("%d-%b-%Y")).upper()
        for contig in self.genome.values():
            contig.annotations = {"molecule_type": "DNA", "date": today_date, "accessions": ""}
            # Leave .id empty to mimic prokka output.
            contig.id = ""

    def rna_calling(self):
        return tools.run_trnascan(self.tmp_input_fna_file)

    def orf_calling(self):
        # Don't call ORFs, just copy it from the input gbk file.
        if self.merge_gbk:
            logger.warning(f" - decouphage will skip this step and use CDS from Genbank file.")
            return cds_calling_from_genbank(self.tmp_input_gbk_file)

        # Use prodigal for ORF calling
        if self.use_prodigal:
            logger.info(f" - with Prodigal")
            return tools.run_prodigal(self.tmp_input_fna_file)

        # Use phanotate for ORF calling
        logger.info(f" - with Phanotate")
        return tools.run_phanotate(self.tmp_input_fna_file)

    def load_features(self):
        for contig_label, orf_list in self.orf_map.items():
            contig = self.genome[contig_label]

            for orf in orf_list:
                idx, start, end, strand = orf.split("_")
                feature = SeqFeature(
                    FeatureLocation(int(start) - 1, int(end)),
                    strand=1 if strand == "+" else -1,
                    type="CDS",
                    qualifiers={},
                )
                contig.features.append(feature)

            for trna in self.trna_map[contig_label]:
                idx, start, end, rna_type, anti_codon = trna.split("_")
                start, end = int(start), int(end)
                strand = 1 if end > start else -1
                positions = [start, end]
                feature = SeqFeature(
                    FeatureLocation(min(positions) - 1, max(positions)),
                    strand=strand,
                    type="tRNA",
                    qualifiers={"product": f"{rna_type}-{anti_codon}"},
                )
                contig.features.append(feature)

        # order features based on start position
        for contig in self.genome.values():
            contig.features.sort(key=lambda x: x.location.start, reverse=False)

        # give feature id an incremental count
        idx = 1
        for contig in self.genome.values():
            for feature in contig.features:
                feature.id = idx
                idx += 1

    def prepare_query_file(self):
        with open(self.tmp_query_faa_file, "a") as fh:
            for contig in self.genome.values():
                for feature in contig.features:
                    if feature.type == "CDS":
                        fh.write(f">{feature.id}\n")
                        fh.write(self.clean_sequence(feature, contig) + "\n")

    def enrich_features(self, qualifiers):
        for contig in self.genome.values():
            for feature in contig.features:
                tag = f"{self.locus_tag}_{int(feature.id):04d}"
                quals = feature.qualifiers
                quals["feature"] = feature.id
                quals["locus_tag"] = tag

                if feature.type == "CDS":
                    blast_result = qualifiers.get(int(feature.id), {})
                    quals["product"] = blast_result.get("stitle", "hypothetical protein"),
                    quals["translation"] = self.clean_sequence(feature, contig)
                    quals["protein_id"] = blast_result.get("sseqid", "N/A")
                feature.qualifiers = quals

    @staticmethod
    def clean_sequence(feature, contig):
        sequence = str(feature.extract(contig.seq).translate(table=11))
        if sequence.endswith("*"):
            sequence = sequence[:-1]
        return sequence



#!/usr/bin/env python3

import logging
import os
import sys
import tarfile
from pathlib import Path
import requests
from alive_progress import alive_bar

logging.basicConfig(level=20)
logger = logging.getLogger(__name__)


def download(file_url: str, file_dst: Path):
    try:

        with file_dst.open('wb') as fh_out, requests.get(file_url, stream=True) as response:
            file_length = response.headers.get('content-length')

            if file_length:
                file_length = int(file_length)
                file_length = int(file_length / 1024)

            with alive_bar(total=file_length) as bar:
                for data in response.iter_content(chunk_size=1024):
                    fh_out.write(data)
                    bar()
    except IOError:
        sys.exit(f'ERROR: Could not download file! url={file_url}, path={file_dst}')


def get_default_config_path() -> Path:
    home_path = os.path.expanduser('~')
    config_path = home_path / Path(".decouphage")
    os.makedirs(config_path, exist_ok=True)
    return config_path


def get_default_database_path(config_path) -> Path:
    database_path = config_path / Path("db")
    os.makedirs(database_path, exist_ok=True)
    return database_path


def run_download():

    default_config_path = get_default_config_path()
    default_database_path = get_default_database_path(default_config_path)
    download_file_path = default_database_path / Path("ncbi_phages.tar.gz")
    logger.info("[1/2] Downloading database...")
    download("https://labs.voorloop.com/archive/decouphage/ncbi_phages_latest.tar.gz", download_file_path)

    logger.info("[2/2] Extracting database...")
    with tarfile.open(download_file_path) as tarball:
        tarball.extractall(default_database_path)

    os.remove(download_file_path)




from collections import defaultdict
from pathlib import Path
from Bio import SeqIO


def cds_calling_from_genbank(input_handle):
    features = defaultdict(list)

    contigs = SeqIO.parse(input_handle, "genbank")
    idx = 1
    for contig in contigs:
        for feature in contig.features:
            if feature.type == "CDS":
                strand = "+" if feature.location.strand == 1 else "-"
                orf = f">{idx}_{feature.location.start+1}_{feature.location.end}_{strand}"
                features[contig.id].append(orf)
                idx += 1
    return features


def get_database_default_path():
    return Path.home() / Path(".decouphage/db/ncbi_phages/ncbi_phages.fa")


def is_fasta(file_path):
    has_sequence = False

    fasta = SeqIO.parse(file_path, "fasta")
    for i in fasta:
        if len(i.seq) > 0:
            has_sequence = True
            return has_sequence
    return has_sequence


def is_genbank(file_path):
    has_sequence = False
    fasta = SeqIO.parse(file_path, "genbank")
    for i in fasta:
        if len(i.seq) > 0:
            has_sequence = True
            return has_sequence
    return has_sequence


def validate_input(input_file):
    try:
        if is_fasta(input_file):
            return "FASTA"
    except UnicodeDecodeError:
        return None

    try:
        if is_genbank(input_file):
            return "GENBANK"
    except UnicodeDecodeError:
        return None



#!/usr/bin/env python3
import json
import logging
import sys

import click
import time
from datetime import timedelta
from src.pipeline import Pipeline
from src.database import run_download

logger = logging.getLogger(__name__)


@click.command()
@click.option('--prodigal', is_flag=True, help='Use prodigal for orf calling instead of phanotate.')
@click.option('-d', '--db', 'database', type=click.Path(exists=True))
@click.option('-o', '--output', 'output', default='output.gbk')
@click.option('-t', '--threads', default=1, show_default=True)
@click.option('--tmpdir', 'tmp_dir', help="Folder for intermediate files.")
@click.option('--no_orf_calling', 'merge_gbk', is_flag=True, help="Annotate CDS from genbank file.")
@click.option('--locus_tag', 'locus_tag', default='dPHG', help="Locus tag prefix.")
@click.option('--download_db', 'download_db', is_flag=True, help="Download default database.")
@click.option('-v', '--verbose', is_flag=True, help="More verbose logging for debugging purpose.")
@click.argument('input_file', type=click.Path(exists=True))
def run_pipeline(prodigal, database, output, input_file, threads, tmp_dir, merge_gbk, locus_tag, download_db, verbose):
    if verbose:
        logging.basicConfig(level=logging.DEBUG)
        logger.setLevel(logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)
        logger.setLevel(logging.INFO)

    if download_db:
        logger.error(" Option --download_db must be used alone.")
        sys.exit(1)

    logger.debug(json.dumps(locals(), indent=4, sort_keys=True))
    start_time = time.monotonic()

    Pipeline(database, input_file, prodigal, threads, output, tmp_dir, merge_gbk, locus_tag)
    end_time = time.monotonic()

    total_time = timedelta(seconds=end_time - start_time)
    logger.info(f"Phage annotation took {total_time}")


if __name__ == '__main__':
    if len(sys.argv) > 1 and sys.argv[1] == "--download_db":
        run_download()
    else:
        run_pipeline()



import setuptools
with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
      name='decouphage',
      version='0.0.3',
      description='decouphage - A tool to annotate phage genomes.',
      long_description=long_description,
      long_description_content_type="text/markdown",
      author='Ricardo Pascal',
      author_email='voorloop@gmail.com',
      url='https://github.com/voorloopnul/decouphage',
      packages=setuptools.find_packages(),
      scripts=['decouphage.py'],
      python_requires='>=3.4',
      install_requires=[
            "biopython >=1.79",
            "phanotate >=1.5.0",
            "pandas >=1.4.2",
            "click >=8.1.3",
            "alive-progress >=2.4.1",
            "requests >=2.27.1",
      ]
)
